@@ -1,6 +1,6 @@
-/* ulogd, Version $Revision: 1.34 $
+/* ulogd, Version $LastChangedRevision: 706 $
  *
- * $Id: ulogd.c,v 1.34 2003/03/05 23:03:49 laforge Exp $
+ * $Id: ulogd.c 706 2005-02-16 11:26:27Z laforge $
  *
  * userspace logging daemon for the iptables ULOG target
  * of the linux 2.4 netfilter subsystem.
@@ -20,7 +20,7 @@
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *
- * $Id: ulogd.c,v 1.34 2003/03/05 23:03:49 laforge Exp $
+ * $Id: ulogd.c 706 2005-02-16 11:26:27Z laforge $
  *
  * Modifications:
  * 	14 Jun 2001 Martin Josefsson <gandalf@wlug.westbo.se>
@@ -29,9 +29,15 @@
  * 	10 Feb 2002 Alessandro Bono <a.bono@libero.it>
  * 		- added support for non-fork mode
  * 		- added support for logging to stdout
+ *
+ * 	09 Sep 2003 Magnus Boden <sarek@ozaba.cx>
+ * 		- added support for more flexible multi-section conffile
+ *
+ * 	20 Apr 2004 Nicolas Pougetoux <nicolas.pougetoux@edelweb.fr>
+ * 		- added suppurt for seteuid()
  */
 
-#define ULOGD_VERSION	"1.00"
+#define ULOGD_VERSION	"1.21"
 
 #include <unistd.h>
 #include <stdio.h>
@@ -45,14 +51,21 @@
 #include <sys/types.h>
 #include <dirent.h>
 #include <getopt.h>
+#include <pwd.h>
+#include <grp.h>
 #include <libipulog/libipulog.h>
 #include <ulogd/conffile.h>
 #include <ulogd/ulogd.h>
 
-/* Size of the netlink receive buffer. If you have _big_ in-kernel
- * queues, you may have to increase this number. 
- * ( --qthreshold 100 * 1500 bytes/packet = 150kB */
-#define ULOGD_BUFSIZE_DEFAULT 65535
+/* Size of the socket recevive memory.  Should be at least the same size as the
+ * 'nlbufsiz' module loadtime parameter of ipt_ULOG.o
+ * If you have _big_ in-kernel queues, you may have to increase this number.  (
+ * --qthreshold 100 * 1500 bytes/packet = 150kB  */
+#define ULOGD_RMEM_DEFAULT	131071
+
+/* Size of the receive buffer for the netlink socket.  Should be at least of
+ * RMEM_DEFAULT size.  */
+#define ULOGD_BUFSIZE_DEFAULT	150000
 
 #ifdef DEBUG
 #define DEBUGP(format, args...) fprintf(stderr, format, ## args)
@@ -476,11 +489,11 @@ static int logfile_open(const char *name)
 }
 
 /* wrapper to handle conffile error codes */
-static int parse_conffile(int final)
+static int parse_conffile(const char *section, config_entry_t *ce)
 {
 	int err;
 
-	err = config_parse_file(final);
+	err = config_parse_file(section, ce);
 
 	switch(err) {
 		case 0:
@@ -493,16 +506,22 @@ static int parse_conffile(int final)
 			break;
 		case -ERRMAND:
 			ulogd_log(ULOGD_ERROR,
-				"mandatory option not found\n");
+				"mandatory option \"%s\" not found\n",
+				config_errce->key);
 			break;
 		case -ERRMULT:
 			ulogd_log(ULOGD_ERROR,
-				"option occurred more than once\n");
+				"option \"%s\" occurred more than once\n",
+				config_errce->key);
 			break;
 		case -ERRUNKN:
 			ulogd_log(ULOGD_ERROR,
-				"unknown config key\n");
-/*				config_errce->key); */
+				"unknown config key \"%s\"\n",
+				config_errce->key);
+			break;
+		case -ERRSECTION:
+			ulogd_log(ULOGD_ERROR,
+				"section \"%s\" not found\n", section);
 			break;
 	}
 	return 1;
@@ -529,26 +548,26 @@ static config_entry_t nlgroup_ce = { &plugin_ce, "nlgroup", CONFIG_TYPE_INT,
 static config_entry_t loglevel_ce = { &nlgroup_ce, "loglevel", CONFIG_TYPE_INT,
 				      CONFIG_OPT_NONE, 0, 
 				      { value: 1 } };
-
-static int init_conffile(char *file)
-{
-	if (config_register_file(file))
-		return 1;
-
-	config_register_key(&loglevel_ce);
-	
-	/* parse config file the first time (for logfile name, ...) */
-	return parse_conffile(0);
-}
+static config_entry_t rmem_ce = { &loglevel_ce, "rmem", CONFIG_TYPE_INT,
+				  CONFIG_OPT_NONE, 0, 
+				  { value: ULOGD_RMEM_DEFAULT } };
 
 static void sigterm_handler(int signal)
 {
+	ulog_output_t *p;
+	
 	ulogd_log(ULOGD_NOTICE, "sigterm received, exiting\n");
 
 	ipulog_destroy_handle(libulog_h);
 	free(libulog_buf);
 	if (logfile != stdout)
 		fclose(logfile);
+
+	for (p = ulogd_outputs; p; p = p->next) {
+		if (p->fini)
+			(*p->fini)();
+	}
+
 	exit(0);
 }
 
@@ -575,13 +594,15 @@ static void print_usage(void)
 {
 	/* FIXME */
 	printf("ulogd Version %s\n", ULOGD_VERSION);
-	printf("Copyright (C) 2000-2003 Harald Welte "
-	       "<laforge@gnumonks.org>\n\n");
-	printf("Paramters:\n");
+	printf("Copyright (C) 2000-2004 Harald Welte "
+	       "<laforge@gnumonks.org>\n");
+	printf("This is free software with ABSOLUTELY NO WARRANTY.\n\n");
+	printf("Parameters:\n");
 	printf("\t-h --help\tThis help page\n");
 	printf("\t-V --version\tPrint version information\n");
 	printf("\t-d --daemon\tDaemonize (fork into background)\n");
 	printf("\t-c --configfile\tUse alternative Configfile\n");
+	printf("\t-u --uid\tChange UID/GID\n");
 }
 
 static struct option opts[] = {
@@ -589,6 +610,7 @@ static struct option opts[] = {
 	{ "daemon", 0, NULL, 'd' },
 	{ "help", 0, NULL, 'h' },
 	{ "configfile", 1, NULL, 'c'},
+	{ "uid", 1, NULL, 'u' },
 	{ 0 }
 };
 
@@ -597,9 +619,16 @@ int main(int argc, char* argv[])
 	int len;
 	int argch;
 	int daemonize = 0;
+	int change_uid = 0;
+	char *user = NULL;
+	struct passwd *pw;
+	uid_t uid = 0;
+	gid_t gid = 0;
 	ulog_packet_msg_t *upkt;
+	ulog_output_t *p;
 
-	while ((argch = getopt_long(argc, argv, "c:dh::V", opts, NULL)) != -1) {
+
+	while ((argch = getopt_long(argc, argv, "c:dh::Vu:", opts, NULL)) != -1) {
 		switch (argch) {
 		default:
 		case '?':
@@ -627,28 +656,33 @@ int main(int argc, char* argv[])
 		case 'c':
 			ulogd_configfile = optarg;
 			break;
+		case 'u':
+			change_uid = 1;
+			user = strdup(optarg);
+			pw = getpwnam(user);
+			if (!pw) {
+				printf("Unknown user %s.\n", user);
+				free(user);
+				exit(1);
+			}
+			uid = pw->pw_uid;
+			gid = pw->pw_gid;
+			break;
 		}
 	}
 
-	if (init_conffile(ulogd_configfile)) {
-		ulogd_log(ULOGD_FATAL, "parse_conffile error\n");
+	if (config_register_file(ulogd_configfile)) {
+		ulogd_log(ULOGD_FATAL, "error registering configfile \"%s\"\n",
+			  ulogd_configfile);
 		exit(1);
 	}
 	
-	logfile_open(logf_ce.u.string);
-
-	/* parse config file the second time (for plugin options) */
-	if (parse_conffile(1)) {
+	/* parse config file */
+	if (parse_conffile("global", &rmem_ce)) {
 		ulogd_log(ULOGD_FATAL, "parse_conffile\n");
 		exit(1);
 	}
 
-#ifdef DEBUG
-	/* dump key and interpreter hash */
-	interh_dump();
-	keyh_dump();
-#endif
-
 	/* allocate a receive buffer */
 	libulog_buf = (unsigned char *) malloc(bufsiz_ce.u.value);
 
@@ -658,10 +692,10 @@ int main(int argc, char* argv[])
 		ipulog_perror(NULL);
 		exit(1);
 	}
-	
+
 	/* create ipulog handle */
-	libulog_h = 
-		ipulog_create_handle(ipulog_group2gmask(nlgroup_ce.u.value));
+	libulog_h = ipulog_create_handle(ipulog_group2gmask(nlgroup_ce.u.value),
+					 rmem_ce.u.value);
 
 	if (!libulog_h) {
 		/* if some error occurrs, print it to stderr */
@@ -670,6 +704,48 @@ int main(int argc, char* argv[])
 		exit(1);
 	}
 
+
+	if (change_uid) {
+		ulogd_log(ULOGD_NOTICE, "Changing UID / GID\n");
+		if (setgid(gid)) {
+			ulogd_log(ULOGD_FATAL, "can't set GID\n");
+			ipulog_perror(NULL);
+			exit(1);
+		}
+		if (setegid(gid)) {
+			ulogd_log(ULOGD_FATAL, "can't sett effective GID\n");
+			ipulog_perror(NULL);
+			exit(1);
+		}
+		if (initgroups(user, gid)) {
+			ulogd_log(ULOGD_FATAL, "can't set user secondary GID\n");
+			ipulog_perror(NULL);
+			exit(1);
+		}
+		if (setuid(uid)) {
+			ulogd_log(ULOGD_FATAL, "can't set UID\n");
+			ipulog_perror(NULL);
+			exit(1);
+		}
+		if (seteuid(uid)) {
+			ulogd_log(ULOGD_FATAL, "can't set effective UID\n");
+			ipulog_perror(NULL);
+			exit(1);
+		}
+	}
+
+	logfile_open(logf_ce.u.string);
+
+	for (p = ulogd_outputs; p; p = p->next) {
+		if (p->init)
+			(*p->init)();
+	}
+
+#ifdef DEBUG
+	/* dump key and interpreter hash */
+	interh_dump();
+	keyh_dump();
+#endif
 	if (daemonize){
 		if (fork()) {
 			exit(0);
@@ -681,8 +757,10 @@ int main(int argc, char* argv[])
 		setsid();
 	}
 
-	signal(SIGTERM, &sigterm_handler);
+	/* send SIGINT to the term handler, since they hit CTRL-C */
+	signal(SIGINT, &sigterm_handler);
 	signal(SIGHUP, &sighup_handler);
+	signal(SIGTERM, &sigterm_handler);
 
 	ulogd_log(ULOGD_NOTICE, 
 		  "initialization finished, entering main loop\n");
